/*
  Modul HTTP
  part of Arduino Mega Server project
*/

#ifdef HTTP_FEATURE

#include <ESP8266WebServer.h>

#define HTTP_PORT 8080
ESP8266WebServer HTTP(HTTP_PORT);

File uploadFile;

void httpInit() {
  initStart("HTTP");
  HTTP.on("/list", HTTP_GET, handleFileList);
  
  HTTP.on("/edit", HTTP_GET, []() {
    if (!handleFileRead("edit.htm")) HTTP.send(404, "text/plain", "FileNotFound");
  });
  
  HTTP.on("/edit", HTTP_PUT,    handleFileCreate);
  HTTP.on("/edit", HTTP_DELETE, handleFileDelete);
  
  //first callback is called after the request has ended with all parsed arguments
  //second callback handles file uploads at that location
  HTTP.on("/edit", HTTP_POST, []() {
    HTTP.send(200, "text/plain", "");
  }, handleFileUpload);

  // called when the url is not defined here, use it to load content from SPIFFS
  HTTP.onNotFound([]() {
    if (!handleFileRead(HTTP.uri()))
      HTTP.send(404, "text/plain", "FileNotFound");
  });
  
  updateInit();
  HTTP.begin();
  timeStamp();
  printValue("HTTP port", String(HTTP_PORT));
  initDone();
} // httpInit()

String getContentType(String filename) {
  if (HTTP.hasArg("download"))         return "application/octet-stream";
  else if (filename.endsWith(".htm"))  return "text/html";
  else if (filename.endsWith(".html")) return "text/html";
  else if (filename.endsWith(".css"))  return "text/css";
  else if (filename.endsWith(".js"))   return "application/javascript";
  else if (filename.endsWith(".png"))  return "image/png";
  else if (filename.endsWith(".gif"))  return "image/gif";
  else if (filename.endsWith(".jpg"))  return "image/jpeg";
  else if (filename.endsWith(".ico"))  return "image/x-icon";
  else if (filename.endsWith(".xml"))  return "text/xml";
  else if (filename.endsWith(".pdf"))  return "application/x-pdf";
  else if (filename.endsWith(".zip"))  return "application/x-zip";
  else if (filename.endsWith(".gz"))   return "application/x-gzip";
                                       return "text/plain";
}

#define BAD_ARGS      "BAD ARGS"
#define BAD_PATH      "BAD PATH"
#define FILE_EXISTS   "FILE EXISTS"
#define CREATE_FAILED "CREATE FAILED"
#define NOT_DIR       "NOT DIR"

void sendOk() {
  HTTP.send(200, "text/plain", "");
}

void sendError(String s) {
  HTTP.send(500, "text/plain", s + "\r\n");
}

bool handleFileRead(String path) {
  if (path.startsWith("/")) {path = path.substring(1, path.length());}
  if (path.endsWith("/")) {path += "index.htm";}
  String contentType = getContentType(path);
  String pathWithGz = path + ".gz";
  if (SD.exists((char *)pathWithGz.c_str()) || SD.exists((char *)path.c_str())) {
    if (SD.exists((char *)pathWithGz.c_str())) {
      path += ".gz";
    }
    File file = SD.open(path.c_str());
    size_t sent = HTTP.streamFile(file, contentType);
    file.close();
    return true;
  }
  return false;
} // handleFileRead( )

void handleFileUpload() {
  if(HTTP.uri() != "/edit") {return;}
  HTTPUpload& upload = HTTP.upload();
  if (upload.status == UPLOAD_FILE_START) {
    if (SD.exists((char *)upload.filename.c_str())) {
      SD.remove((char *)upload.filename.c_str());
    }
    uploadFile = SD.open(upload.filename.c_str(), FILE_WRITE);
  }
  else if(upload.status == UPLOAD_FILE_WRITE){
    if (uploadFile) {uploadFile.write(upload.buf, upload.currentSize);}
  }
  else if(upload.status == UPLOAD_FILE_END){
    if (uploadFile) {uploadFile.close();}
  }
} // handleFileUpload()

void handleFileDelete() {
  if (HTTP.args() == 0) {sendError(BAD_ARGS); return;}
  String path = HTTP.arg(0);
  if (path.startsWith("/")) {path = path.substring(1, path.length());}
  Serial.println("handleFileDelete: " + path);
  if (path == "/") {sendError(BAD_PATH); return;}
  if (!SD.exists((char *)path.c_str())) {
    return HTTP.send(404, "text/plain", "FileNotFound: " + path);
  }
  SD.remove((char *)path.c_str());
  HTTP.send(200, "text/plain", "");
  path = String();
} // handleFileDelete()

void handleFileCreate() {
  if (HTTP.args() == 0) {sendError(BAD_ARGS); return;}
  String path = HTTP.arg(0);
  if (path == "/") {sendError(BAD_PATH); return;}
  if (SD.exists((char *)path.c_str())) {sendError(FILE_EXISTS); return;}
  File file = SD.open(path.c_str(), FILE_WRITE);
  if (file) {
    file.close();
  } else {
      sendError(CREATE_FAILED);
      return;
    }
  sendOk();
  path = String();
} // handleFileCreate();

void handleFileList() {
  if (!HTTP.hasArg("dir")) {sendError(BAD_ARGS); return;}
  String path = HTTP.arg("dir");
  if (path != "/" && !SD.exists((char *)path.c_str())) {sendError(BAD_PATH); return;}
  File dir = SD.open((char *)path.c_str());
  path = String();
  if (!dir.isDirectory()) {
    dir.close();
    sendError(NOT_DIR);
    return;
  }
  dir.rewindDirectory();
  HTTP.setContentLength(CONTENT_LENGTH_UNKNOWN);
  WiFiClient client = HTTP.client();

  String output = "[";
  for (int i = 0; true; ++i) {
    File entry = dir.openNextFile();
    if (!entry) {break;}
    if (i > 0) {output += ',';}
    output += "{\"type\":\"";
    output += (entry.isDirectory()) ? "dir" : "file";
    output += "\",\"name\":\"";
    output += entry.name();
    output += "\"";
    output += "}";
    entry.close();
  }
  output += "]";
  HTTP.send(200, "text/json", output);
  dir.close();
} // handleFileList()

void updateInit() {
  HTTP.on("/update", HTTP_POST, []() {
    HTTP.sendHeader("Connection", "close");
    HTTP.sendHeader("Access-Control-Allow-Origin", "*");
    HTTP.send(200, "text/plain", (Update.hasError()) ? "FAIL" : "OK");
    ESP.restart();
  }, []() {
    HTTPUpload& upload = HTTP.upload();
    if (upload.status == UPLOAD_FILE_START) {
      Serial.setDebugOutput(true);
      WiFiUDP::stopAll();
      Serial.printf("\n=== UPDATE PROCEDURE ==========\n%s\n", upload.filename.c_str());
      uint32_t maxSketchSpace = (ESP.getFreeSketchSpace() - 0x1000) & 0xFFFFF000;
      if (!Update.begin(maxSketchSpace)) { //start with max available size
        Update.printError(Serial);
      }
    }
    else if (upload.status == UPLOAD_FILE_WRITE) {
      if (Update.write(upload.buf, upload.currentSize) != upload.currentSize) {
        Update.printError(Serial);
      }
    } 
    else if (upload.status == UPLOAD_FILE_END) {
      if (Update.end(true)) { //true to set the size to the current progress
        Serial.printf("Update Success: %u\n=== REBOOTING =================\n", upload.totalSize);
      } else {
          Update.printError(Serial);
        }
      Serial.setDebugOutput(false);
    }
    yield();
  });
} // updateInit()

void httpWork() {
  HTTP.handleClient();
  delay(1);
}

#endif // HTTP_FEATURE
